# ============================================================================
# ============================================================================
# Copyright (c) 2020 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: nghia.vo@diamond.ac.uk
# Description: Python implementations of preprocessing techniques for
# tomography.
# Publication date: 10th July 2020
# ============================================================================

"""
Module of correction methods in the preprocessing stage:
- Flat-field correction.
- Distortion correction.
- MTF deconvolution.
"""

import numpy as np
from scipy.ndimage import map_coordinates
import pyfftw.interfaces.scipy_fftpack as fft
from algotom.preprocessing.removal import *
from algotom.preprocessing.filtering import *


def flat_field_correction(proj, flat, dark, ratio=1.0, use_dark=True, **options):
    """
    Do flat-field correction with options to remove zinger arifacts and/or
    stripe artifacts.

    Parameters
    ----------
    proj : array_like
        3D or 2D array. Projection images or a sinogram image.
    flat : array_like
        2D or 1D array. Flat-field image or a single row of it.
    dark : array_like
        2D or 1D array. Dark-field image or a single row of it.
    ratio : float
        Ratio between exposure time used for recording projections
        and exposure time used for recording flat field.
    use_dark : bool
        Subtracting dark field if True. May no need in some cases.
    options : dict, optional
        Apply a zinger removal method and/or ring removal methods.
        E.g option1={"method": "dezinger", "para1": 0.001, "para2": 1},
        option2={"method": "remove_stripe_based_sorting", "para1": 15, "para2": 1}

    Returns
    -------
    ndarray
        3D or 2D array. Corrected projections or corrected sinograms.
    """
    flat = ratio * flat
    if use_dark:
        flatdark = flat - dark
        try:
            proj_corr = (np.float32(proj) - dark) / flatdark
        except ZeroDivisionError:
            nmean = np.mean(flatdark)
            if nmean != 0.0:
                flatdark[flatdark == 0.0] = nmean
            else:
                flatdark[flatdark == 0.0] = 1
            proj_corr = (np.float32(proj) - dark) / flatdark
    else:
        try:
            proj_corr = np.float32(proj) / flat
        except ZeroDivisionError:
            nmean = np.mean(flat)
            if nmean != 0.0:
                flat[flat == 0.0] = nmean
            else:
                flat[flat == 0.0] = 1
            proj_corr = np.float32(proj) / flat
    if len(options) != 0:
        for opt_name in options:
            opt = options[opt_name]
            if 'method' in opt.keys():
                method = opt['method']
                list_para = tuple(opt.values())[1:]
                if proj_corr.ndim == 2:
                    proj_corr = globals()[method](proj_corr, *list_para)
                else:
                    for i in np.arange(proj_corr.shape[1]):
                        proj_corr[:, i, :] = globals()[method](proj_corr[:, i, :], *list_para)
            else:
                raise ValueError("Incorrect option: {}".format(opt))
    return proj_corr


def unwarp_projection(proj, xcenter, ycenter, list_fact):
    """
    Apply distortion correction to a projection image using the polynomial
    backward model [1].

    Parameters
    ----------
    proj : array_like
        2D array. Projection image.
    xcenter : float
        Center of distortion in x-direction.
    ycenter : float
        Center of distortion in y-direction.
    list_fact : list of float
        Polynomial coefficients of the backward model.

    Returns
    -------
    ndarray
        2D array. Distortion corrected.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.23.032859

    """
    (height, width) = proj.shape
    xu_list = np.arange(width) - xcenter
    yu_list = np.arange(height) - ycenter
    xu_mat, yu_mat = np.meshgrid(xu_list, yu_list)
    ru_mat = np.sqrt(xu_mat ** 2 + yu_mat ** 2)
    fact_mat = np.sum(np.asarray(
        [factor * ru_mat ** i for i, factor in enumerate(list_fact)]), axis=0)
    xd_mat = np.float32(np.clip(xcenter + fact_mat * xu_mat, 0, width - 1))
    yd_mat = np.float32(np.clip(ycenter + fact_mat * yu_mat, 0, height - 1))
    indices = np.reshape(yd_mat, (-1, 1)), np.reshape(xd_mat, (-1, 1))
    proj = map_coordinates(proj, indices, order=1, mode='reflect')
    return proj.reshape((height, width))


def unwarp_sinogram(data, index, xcenter, ycenter, list_fact, **option):
    """
    Unwarp sinogram [:,index.:] of a 3D tomographic dataset or
    a hdf/nxs object.

    Parameters
    ----------
    data : array_like or hdf object
        3D array.
    index : int
        Index of the sinogram.
    xcenter : float
        Center of distortion in x-direction.
    ycenter : float
        Center of distortion in y-direction.
    list_fact : list of float
        Polynomial coefficients of the backward model.
    option : list or tuple of int
        To extract subset data along axis 0 from a hdf object. E.g option =
        (start, stop, step)

    Returns
    -------
    ndarray
        2D array. Distortion-corrected sinogram.
    """
    if data.ndim < 3:
        raise ValueError("Input must be a 3D data !!!")
    (depth, height, width) = data.shape
    if len(option) != 0:
        opt = option[list(option.keys())[0]]
        start, stop, step = opt
    else:
        start, stop, step = 0, depth, 1
    list_idx = range(start, stop, step)
    depth = len(list_idx)
    xu_list = np.arange(0, width) - xcenter
    yu = index - ycenter
    ru_list = np.sqrt(xu_list ** 2 + yu ** 2)
    flist = np.sum(np.asarray(
        [factor * ru_list ** i for i, factor in enumerate(list_fact)]), axis=0)
    xd_list = np.clip(xcenter + flist * xu_list, 0, width - 1)
    yd_list = np.clip(ycenter + flist * yu, 0, height - 1)
    yd_min = np.int16(np.floor(np.amin(yd_list)))
    yd_max = np.int16(np.ceil(np.amax(yd_list))) + 1
    yd_list = yd_list - yd_min
    indices = yd_list, xd_list
    sinogram = np.zeros((depth, width), dtype=np.float32)
    for i, idx in enumerate(list_idx):
        sinogram[i] = map_coordinates(
            data[idx, yd_min:yd_max, :], indices, order=1, mode='reflect')
    return sinogram


def mapping(mat, xmat, ymat):
    """
    Apply a geometric transformation to a 2D array

    Parameters
    ----------
    mat : array_like
        2D array.
    xmat : array_like
        2D array of the x-coordinates.
    ymat : array_like
        2D array of the y-coordinates.

    Returns
    -------
    ndarray
        2D array.
    """
    indices = np.reshape(ymat, (-1, 1)), np.reshape(xmat, (-1, 1))
    mat = map_coordinates(mat, indices, order=1, mode='reflect')
    return mat.reshape(xmat.shape)


def unwarp_sinogram_chunk(data, start_index, stop_index, xcenter, ycenter, list_fact, **option):
    """
    Unwarp chunk of sinograms [:, start_index: stop_index, :]
    of a 3D tomographic dataset or a hdf/nxs object.

    Parameters
    ----------
    data : array_like or hdf object
        3D array.
    start_index : int
        Starting index of the slice.
    stop_index : int
        Stopping index of the slice.
    xcenter : float
        Center of distortion in x-direction.
    ycenter : float
        Center of distortion in y-direction.
    list_fact : list of float
        Polynomial coefficients of the backward model.
    option : list or tuple of int
        To extract subset data along axis 0 from a hdf object. E.g option =
        [start, stop, step]

    Returns
    -------
    ndarray
        3D array. Distortion corrected.
    """
    if data.ndim < 3:
        raise ValueError("Input must be a 3D data !!!")
    (depth, height, width) = data.shape
    if len(option) != 0:
        opt = option[list(option.keys())[0]]
        start, stop, step = opt
    else:
        start, stop, step = 0, depth, 1
    list_idx = range(start, stop, step)
    if stop_index == -1:
        stop_index = height
    xu_list = np.arange(0, width) - xcenter
    yu1 = start_index - ycenter
    ru_list = np.sqrt(xu_list ** 2 + yu1 ** 2)
    flist = np.sum(np.asarray(
        [factor * ru_list ** i for i, factor in enumerate(list_fact)]), axis=0)
    yd_list1 = np.clip(ycenter + flist * yu1, 0, height - 1)
    yu2 = stop_index - ycenter
    ru_list = np.sqrt(xu_list ** 2 + yu2 ** 2)
    flist = np.sum(np.asarray(
        [factor * ru_list ** i for i, factor in enumerate(list_fact)]), axis=0)
    yd_list2 = np.clip(ycenter + flist * yu2, 0, height - 1)
    yd_min = np.int16(np.floor(np.amin(yd_list1)))
    yd_max = np.int16(np.ceil(np.amax(yd_list2)))
    yu_list = np.arange(start_index, stop_index) - ycenter
    xu_mat, yu_mat = np.meshgrid(xu_list, yu_list)
    ru_mat = np.sqrt(xu_mat ** 2 + yu_mat ** 2)
    fact_mat = np.sum(np.asarray(
        [factor * ru_mat ** i for i, factor in enumerate(list_fact)]), axis=0)
    xd_mat = np.float32(np.clip(xcenter + fact_mat * xu_mat, 0, width - 1))
    yd_mat = np.float32(
        np.clip(ycenter + fact_mat * yu_mat, 0, height - 1)) - yd_min
    sino_chunk = np.asarray(
        [mapping(data[i, yd_min:yd_max, :], xd_mat, yd_mat) for i in list_idx])
    return sino_chunk


def mtf_deconvolution(mat, window, pad):
    """
    Deconvolve an projection image using division in the Fourier domain.
    Window can be determined using the approach in [1].

    Parameters
    ----------
    mat : array_like
        2D array. Projection image
    window : array_like
        2D array. MTF function
    pad : int
        Padding width to reduce the side effects of the Fourier transform.

    Returns
    -------
    ndarray
        2D array. Deconvolved image.

    References
    ----------
    .. [1] https://doi.org/10.1117/12.2530324
    """
    (nrow1, ncol1) = mat.shape
    (nrow2, ncol2) = window.shape
    if (nrow1 > nrow2) or (ncol1 > ncol2):
        raise ValueError(
            "The sizes of the image are larger than the sizes of the window")
    else:
        pad_row = nrow2 - nrow1
        pad_col = ncol2 - ncol1
        mat = np.pad(mat, ((0, pad_row), (0, pad_col)), mode="edge")
    mat_pad = np.pad(mat, pad, mode="reflect")
    win_pad = np.pad(window, pad, mode="constant", constant_values=1.0)
    mat_dec = fft.ifft2(fft.fft2(mat_pad) / fft.ifftshift(win_pad))
    mat_dec = mat_dec[pad: nrow2 + pad, pad: ncol2 + pad]
    return mat_dec[:nrow1, :ncol1]
