# ============================================================================
# ============================================================================
# Copyright (c) 2020 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: nghia.vo@diamond.ac.uk
# Description: Python implementations of preprocessing techniques for
# tomography.
# Publication date: 10th July 2020
# ============================================================================

"""
Module of filtering methods in the preprocessing stage:
- Fresnel filter.
- Converting a 180-degree sinogram to a 360-degree sinogram.
- Double-wedge filter.
"""

import numpy as np
import pyfftw.interfaces.scipy_fftpack as fft
from scipy.ndimage import shift


def make_fresnel_window(height, width, ratio, dim):
    """
    Create a low pass window based on the Fresnel propagator.
    It is used to denoise a projection image (dim=2) or a
    sinogram image (dim=1).

    Parameters
    ----------
    height : int
        Image height
    width : int
        Image width
    ratio : float
        To define the shape of the window.
    dim : {1, 2}
        Use "1" if working on a sinogram image and "2" if working on
        a projection image.

    Returns
    -------
    array_like
        2D array.
    """
    ycenter = int(np.ceil((height - 1) * 0.5))
    xcenter = int(np.ceil((width - 1) * 0.5))
    if dim == 2:
        ulist = (1.0 * np.arange(0, width) - xcenter) / width
        vlist = (1.0 * np.arange(0, height) - ycenter) / height
        u, v = np.meshgrid(ulist, vlist)
        window = 1.0 + ratio * (u ** 2 + v ** 2)
    else:
        ulist = (1.0 * np.arange(0, width) - xcenter) / width
        win1d = 1.0 + ratio * ulist ** 2
        window = np.tile(win1d, (height, 1))
    return window


def fresnel_filter(mat, ratio, dim, window=None, pad=150):
    """
    Apply a low-pass filter based on the Fresnel propagator to an image [1].

    Parameters
    ----------
    mat : array_like
        2D array. Projection image or sinogram image.
    ratio : float
        Define the shape of the window. Larger is more smoothing.
    dim : {1, 2}
        Use "1" if working on a sinogram image and "2" if working on
        a projection image.
    window : array_like, optional
        Window for deconvolution.
    pad : int
        Padding width.

    Returns
    -------
    array_like
        2D array. Filtered image.

    References
    ----------
    .. [1] N. T. Vo, R.C Atwood, M. Drakopoulos, and T. Connolley, "Data
    processing methods and data acquisition for samples larger than the
    field of view in parallel-beam tomography," Optics express, submitted.
    """
    (nrow, ncol) = mat.shape
    if dim == 2:
        if window is None:
            window = make_fresnel_window(nrow, ncol, ratio, dim)
        mat_pad = np.pad(mat, pad, mode="edge")
        win_pad = np.pad(window, pad, mode="edge")
        mat_dec = fft.ifft2(fft.fft2(-np.log(mat_pad)) / fft.ifftshift(win_pad))
        mat_dec = np.abs(mat_dec[pad:pad + nrow, pad:pad + ncol])
    else:
        if window is None:
            window = make_fresnel_window(nrow, ncol, ratio, dim)
        mat_pad = np.pad(-np.log(mat), ((0, 0), (pad, pad)), mode='edge')
        win_pad = np.pad(window, ((0, 0), (pad, pad)), mode="edge")
        mat_fft = np.fft.fftshift(fft.fft(mat_pad), axes=1) / win_pad
        mat_dec = fft.ifft(np.fft.ifftshift(mat_fft, axes=1))
        mat_dec = np.abs(mat_dec[:, pad:pad + ncol])
    return np.float32(np.exp(-mat_dec))


def make_double_wedge_mask(height, width, radius):
    """
    Generate a double-wedge binary mask using Eq. (3) in [1].
    Values outside the double-wedge region correspond to 0.0s.

    Parameters
    ----------
    height : int
        Image height.
    width : int
        Image width.
    radius : int
        Radius of an object, in pixel unit.

    Returns
    -------
    ndarray
        2D binary mask.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.22.019078
    """
    du = 1.0 / width
    dv = (height - 1.0) / (height * 2.0 * np.pi)
    ndrop = 1
    ycenter = np.int16(np.ceil((height - 1) / 2.0))
    xcenter = np.int16(np.ceil((width - 1) / 2.0))
    mask = np.ones((height, width), dtype=np.float32)
    for i in range(height):
        num = np.int16(np.ceil(((i - ycenter) * dv / radius) / du))
        (p1, p2) = np.int16(np.clip(
            np.sort((-num + xcenter, num + xcenter)), 0, width - 1))
        mask[i, p1:p2 + 1] = 0.0
    mask[ycenter - ndrop:ycenter + ndrop + 1, :] = 1.0
    return mask


def convert_sinogram_180_to_360(sino_180, center):
    """
    Convert a 180-degree sinogram to a 360-degree sinogram [1].

    Parameters
    ----------
    sino_180 : array_like
        2D array. 180-degree sinogram.
    center : float
        Center-of-rotation.

    Returns
    -------
    array_like
        360-degree sinogram doubled in height compared to the input.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.22.019078
    """
    (_, ncol) = sino_180.shape
    xcenter = (ncol - 1.0) / 2.0
    shift_x = xcenter - center
    sino_flip = shift(
        np.fliplr(shift(sino_180, (0, shift_x), mode='nearest'))
        , (0, -shift_x), mode='nearest')
    return np.vstack((sino_180, sino_flip))


def double_wedge_filter(sinogram, center=0, sino_type="180",
                        iteration=1, mask=None, ratio=1.0):
    """
    Apply double-wedge filter to a sinogram image [1].

    Parameters
    ----------
    sinogram : array_like
        2D array. 180-degree sinogram or 360-degree sinogram.
    center : float, optional
        Center-of-rotation. No need for a 360-sinogram.
    sino_type : {"180", "360"}
        Sinogram type : 180-degree or 360-degree.
    iteration : int
        Number of iteration.
    mask : array_like, optional
        Double-wedge binary mask.
    ratio : float, optional
        Define the cut-off angle of the double-wedge filter.

    Returns
    -------
    array_like
        2D array. Filtered sinogram.

    References
    ----------
    .. [1] N. T. Vo, R.C Atwood, M. Drakopoulos, and T. Connolley, "Data
    processing methods and data acquisition for samples larger than the
    field of view in parallel-beam tomography," Optics express, submitted.
    """
    if sino_type == "180":
        if center == 0:
            raise ValueError(
                "Please provide the location of the rotation axis")
        sinogram = convert_sinogram_180_to_360(sinogram, center)
    (nrow, ncol) = sinogram.shape
    if mask is None:
        mask = make_double_wedge_mask(nrow, ncol, ratio * ncol / 2.0)
    else:
        if mask.shape != sinogram.shape:
            raise ValueError(
                "Shape of the sinogram {0} and the mask "
                "{1} is not the same!!!".format(sinogram.shape, mask.shape))
    sino_filt = np.copy(sinogram)
    pad = ncol // 3
    mask_pad = np.pad(mask, ((0, 0), (pad, pad)), mode="edge")
    for i in range(iteration):
        sino_filt = np.pad(sino_filt, ((0, 0), (pad, pad)), mode="edge")
        sino_filt = np.real(
            fft.ifft2(fft.ifftshift(fft.fftshift(fft.fft2(sino_filt)) * mask_pad)))
        sino_filt = sino_filt[:, pad:ncol + pad]
    if sino_type == "180":
        sino_filt = sino_filt[:nrow // 2]
    return sino_filt
