# ============================================================================
# ============================================================================
# Copyright (c) 2020 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: nghia.vo@diamond.ac.uk
# Description: Python implementations of preprocessing techniques for
# tomography.
# Publication date: 10th July 2020
# ============================================================================

"""
Module of correction methods in the preprocessing stage:
- Stripe artifact removal.
- Zinger removal.
- Blob removal.
"""

import numpy as np
from scipy.signal.windows import gaussian
from scipy.signal import savgol_filter
from scipy.ndimage import median_filter
from scipy.ndimage import gaussian_filter
from scipy.ndimage import binary_dilation
from scipy.ndimage import uniform_filter1d
from scipy import interpolate
import pyfftw.interfaces.scipy_fftpack as fft
import pywt


def remove_stripe_based_sorting(sinogram, size, dim=1):
    """
    Remove stripe artifacts in a sinogram using the sorting technique,
    algorithm 3 in [1]. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    size : int
        Window size of the median filter.
    dim : {1, 2}, optional
        Dimension of the window.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    sinogram = np.transpose(sinogram)
    (nrow, ncol) = sinogram.shape
    list_index = np.arange(0.0, ncol, 1.0)
    mat_index = np.tile(list_index, (nrow, 1))
    mat_comb = np.asarray(np.dstack((mat_index, sinogram)))
    mat_sort = np.asarray(
        [row[row[:, 1].argsort()] for row in mat_comb])
    if dim == 2:
        mat_sort[:, :, 1] = median_filter(mat_sort[:, :, 1], (size, size))
    else:
        mat_sort[:, :, 1] = median_filter(mat_sort[:, :, 1], (size, 1))
    mat_sort_back = np.asarray(
        [row[row[:, 0].argsort()] for row in mat_sort])
    return np.transpose(mat_sort_back[:, :, 1])


def remove_stripe_based_filtering(sinogram, sigma, size, dim=1):
    """
    Remove stripe artifacts in a sinogram using the filtering technique,
    algorithm 2 in [1]. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image
    sigma : int
        Sigma of the Gaussian window used to separate the low-pass and
        high-pass components of the intensity profile of each column.
    size : int
        Window size of the median filter.
    dim : {1, 2}, optional
        Dimension of the window.

    Returns
    -------
    array_like
         2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    pad = min(150, int(0.1 * sinogram.shape[0]))
    sinogram = np.transpose(sinogram)
    sino_pad = np.pad(sinogram, ((0, 0), (pad, pad)), mode='reflect')
    (_, ncol) = sino_pad.shape
    window = gaussian(ncol, std=sigma)
    list_sign = np.power(-1.0, np.arange(ncol))
    sino_smooth = np.copy(sinogram)
    for i, sino_1d in enumerate(sino_pad):
        sino_smooth[i] = np.real(
            fft.ifft(fft.fft(sino_1d * list_sign) * window) * list_sign)[pad:ncol - pad]
    sino_sharp = sinogram - sino_smooth
    if dim == 2:
        sino_smooth_cor = median_filter(sino_smooth, (size, size))
    else:
        sino_smooth_cor = median_filter(sino_smooth, (size, 1))
    return np.transpose(sino_smooth_cor + sino_sharp)


def make_2d_gaussian_window(height, width, sigmax, sigmay):
    """
    Create a 2D Gaussian window.

    Parameters
    ----------
    height : int
        Height of the image.
    width : int
        Width of the image.
    sigmax : int
        Sigma in the x-direction.
    sigmay : int
        Sigma in the y-direction.

    Returns
    -------
    ndarray
        2D array.
    """
    xcenter = (width - 1.0) / 2.0
    ycenter = (height - 1.0) / 2.0
    y, x = np.ogrid[-ycenter:height - ycenter, -xcenter:width - xcenter]
    window = np.exp(-(x ** 2 / (2 * sigmax ** 2) + y ** 2 / (2 * sigmay ** 2)))
    return window


def apply_gaussian_filter(mat, sigmax, sigmay, pad):
    """
    Filtering an image using a 2D Gaussian window.

    Parameters
    ----------
    mat : array_like
        2D array.
    sigmax : int
        Sigma in the x-direction.
    sigmay : int
        Sigma in the y-direction.
    pad : int
        Padding for the Fourier transform.

    Returns
    -------
    ndarray
        2D array. Filtered image.
    """
    mat_pad = np.pad(mat, ((0, 0), (pad, pad)), mode='edge')
    mat_pad = np.pad(mat_pad, ((pad, pad), (0, 0)), mode='mean')
    (nrow, ncol) = mat_pad.shape
    window = make_2d_gaussian_window(nrow, ncol, sigmax, sigmay)
    listx = np.arange(0, ncol)
    listy = np.arange(0, nrow)
    x, y = np.meshgrid(listx, listy)
    mat_sign = np.power(-1.0, x + y)
    mat_filt = np.real(
        fft.ifft2(fft.fft2(mat_pad * mat_sign) * window) * mat_sign)
    return mat_filt[pad:nrow - pad, pad:ncol - pad]


def remove_stripe_based_fitting(sinogram, order, sigmax, sigmay):
    """
    Remove stripes using the fitting technique, algorithm 1 in [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image
    order : int
        Polynomial fit order.
    sigmax : int
        Sigma of the Gaussian window in the x-direction.
    sigmay : int
        Sigma of the Gaussian window in the y-direction.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    (nrow, _) = sinogram.shape
    pad = min(150, int(0.1 * sinogram.shape[0]))
    nrow1 = nrow
    if nrow1 % 2 == 0:
        nrow1 = nrow1 - 1
    sino_fit = savgol_filter(sinogram, nrow1, order, axis=0, mode='mirror')
    sino_fit_smooth = apply_gaussian_filter(sino_fit, sigmax, sigmay, pad)
    num1 = np.mean(sino_fit)
    num2 = np.mean(sino_fit_smooth)
    sino_fit_smooth = num1 * sino_fit_smooth / num2
    return (sinogram / sino_fit) * sino_fit_smooth


def detect_stripe(list_data, snr):
    """
    Locate stripe positions using Algorithm 4 in [1]

    Parameters
    ----------
    list_data : array_like
        1D array. Normalized data.
    snr : float
        Ratio used to segment between useful information and noise.

    Returns
    -------
    ndarray
        1D binary mask.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    npoint = len(list_data)
    list_sort = np.sort(list_data)
    listx = np.arange(0, npoint, 1.0)
    ndrop = np.int16(0.25 * npoint)
    (slope, intercept) = np.polyfit(listx[ndrop:-ndrop - 1], list_sort[ndrop:-ndrop - 1], 1)
    y_end = intercept + slope * listx[-1]
    noise_level = np.abs(y_end - intercept)
    val1 = np.abs(list_sort[-1] - y_end) / noise_level
    val2 = np.abs(intercept - list_sort[0]) / noise_level
    list_mask = np.zeros(npoint, dtype=np.float32)
    if val1 >= snr:
        upper_thresh = y_end + noise_level * snr * 0.5
        list_mask[list_data > upper_thresh] = 1.0
    if val2 >= snr:
        lower_thresh = intercept - noise_level * snr * 0.5
        list_mask[list_data <= lower_thresh] = 1.0
    return list_mask


def remove_large_stripe(sinogram, snr, size, drop_ratio=0.1):
    """
    Remove large stripes, algorithm 5 in [1], by: locating stripes, normalizing
    to remove full stripes, and using the sorting technique [1] to remove partial
    stripes. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image
    snr : float
        Ratio used to segment between useful information and noise.
    size : int
        Window size of the median filter.
    drop_ratio : float, optional
        Ratio of pixels to be dropped, which is used to to reduce
        the possibility of the false detection of stripes.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    drop_ratio = np.clip(drop_ratio, 0.0, 0.8)
    (nrow, ncol) = sinogram.shape
    ndrop = int(0.5 * drop_ratio * nrow)
    sino_sort = np.sort(sinogram, axis=0)
    sino_smooth = median_filter(sino_sort, (1, size))
    list1 = np.mean(sino_sort[ndrop:nrow - ndrop], axis=0)
    list2 = np.mean(sino_smooth[ndrop:nrow - ndrop], axis=0)
    list2[list2 == 0.0] = np.mean(list2)
    list_fact = list1 / list2
    list_mask = detect_stripe(list_fact, snr)
    list_mask = np.float32(binary_dilation(list_mask, iterations=1))
    mat_fact = np.tile(list_fact, (nrow, 1))
    sinogram = sinogram / mat_fact  # Normalization
    sino_tran = np.transpose(sinogram)
    list_index = np.arange(0.0, nrow, 1.0)
    mat_index = np.tile(list_index, (ncol, 1))
    mat_comb = np.asarray(np.dstack((mat_index, sino_tran)))
    mat_sort = np.asarray(
        [row[row[:, 1].argsort()] for row in mat_comb])
    mat_sort[:, :, 1] = np.transpose(sino_smooth)
    mat_sort_back = np.asarray(
        [row[row[:, 0].argsort()] for row in mat_sort])
    sino_cor = np.transpose(mat_sort_back[:, :, 1])
    listx_miss = np.where(list_mask > 0.0)[0]
    sinogram[:, listx_miss] = sino_cor[:, listx_miss]
    return sinogram


def remove_unresponsive_and_fluctuating_stripe(sinogram, snr, size, residual=False):
    """
    Remove unresponsive or fluctuating stripes, algorithm 6 in [1], by:
    locating stripes, correcting using interpolation. Angular direction
    is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    snr : float
        Ratio used to segment between useful information and noise
    size : int
        Window size of the median filter.
    residual : bool, optional
        Removing residual stripes if True.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    sinogram = np.copy(sinogram)  # Make it mutable
    (nrow, _) = sinogram.shape
    sino_smooth = np.apply_along_axis(uniform_filter1d, 0, sinogram, 10)
    list_diff = np.sum(np.abs(sinogram - sino_smooth), axis=0)
    list_diff_bck = median_filter(list_diff, size)
    nmean = np.mean(np.abs(list_diff_bck))
    list_diff_bck[list_diff_bck == 0.0] = nmean
    list_fact = list_diff / list_diff_bck
    list_mask = detect_stripe(list_fact, snr)
    list_mask = np.float32(binary_dilation(list_mask, iterations=1))
    list_mask[0:2] = 0.0
    list_mask[-2:] = 0.0
    listx = np.where(list_mask < 1.0)[0]
    listy = np.arange(nrow)
    mat = sinogram[:, listx]
    finter = interpolate.interp2d(listx, listy, mat, kind='linear')
    listx_miss = np.where(list_mask > 0.0)[0]
    if len(listx_miss) > 0:
        sinogram[:, listx_miss] = finter(listx_miss, listy)
    if residual is True:
        sinogram = remove_large_stripe(sinogram, snr, size)
    return sinogram


def remove_all_stripe(sinogram, snr, la_size, sm_size, drop_ratio=0.1, dim=1):
    """
    Remove all types of stripe artifacts by combining algorithm 6, 5, 4,
    and 3 in [1]. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    snr : float
        Ratio used to segment between useful information and noise.
    la_size : int
        Window size of the median filter to remove large stripes.
    sm_size : int
        Window size of the median filter to remove small-to-medium stripes.
    drop_ratio : float, optional
        Ratio of pixels to be dropped, which is used to to reduce
        the possibility of the false detection of stripes.
    dim : {1, 2}, optional
        Dimension of the window.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    sinogram = remove_unresponsive_and_fluctuating_stripe(
        sinogram, snr, la_size)
    sinogram = remove_large_stripe(sinogram, snr, la_size, drop_ratio)
    sinogram = remove_stripe_based_sorting(sinogram, sm_size, dim)
    return sinogram


def remove_stripe_based_filtering_sorting(sinogram, sigma, size, dim=1):
    """
    Removing stripes using the filtering and sorting technique, combination of
    algorithm 2 and algorithm 3 in [1]. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    sigma : int
        Sigma of the Gaussian window used to separate the low-pass and
        high-pass components of the intensity profile of each column.
    size : int
        Window size of the median filter.
    dim : {1, 2}, optional
        Dimension of the window.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    pad = min(150, int(0.1 * sinogram.shape[0]))
    sinogram = np.transpose(sinogram)
    sino_pad = np.pad(sinogram, ((0, 0), (pad, pad)), mode='reflect')
    (_, ncol) = sino_pad.shape
    window = gaussian(ncol, std=sigma)
    list_sign = np.power(-1.0, np.arange(ncol))
    sino_smooth = np.copy(sinogram)
    for i, sino_1d in enumerate(sino_pad):
        sino_smooth[i] = np.real(
            fft.ifft(fft.fft(sino_1d * list_sign) * window) * list_sign)[pad:ncol - pad]
    sino_sharp = sinogram - sino_smooth
    sino_smooth_cor = np.transpose(
        remove_stripe_based_sorting(np.transpose(sino_smooth), size, dim))
    return np.transpose(sino_smooth_cor + sino_sharp)


def remove_stripe_based_sorting_fitting(sinogram, order, sigmax, sigmay):
    """
    Remove stripes using the sorting and fitting technique, combination of
    algorithm 2 and 1 in [1]. Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    order : int
        Polynomial fit order.
    sigmax : int
        Sigma of the Gaussian window in the x-direction.
    sigmay : int
        Sigma of the Gaussian window in the y-direction.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    sinogram = np.transpose(sinogram)
    (nrow, ncol) = sinogram.shape
    list_index = np.arange(0.0, ncol, 1.0)
    mat_index = np.tile(list_index, (nrow, 1))
    mat_comb = np.asarray(np.dstack((mat_index, sinogram)))
    mat_sort = np.asarray(
        [row[row[:, 1].argsort()] for row in mat_comb])
    sino_sort = mat_sort[:, :, 1]
    sino_cor = np.transpose(
        remove_stripe_based_fitting(np.transpose(sino_sort), order, sigmax, sigmay))
    mat_sort[:, :, 1] = sino_cor
    mat_sort_back = np.asarray(
        [row[row[:, 0].argsort()] for row in mat_sort])
    return np.transpose(mat_sort_back[:, :, 1])


def remove_stripe_based_2d_filtering_sorting(sinogram, sigma, size, dim=1):
    """
    Remove stripes using a 2D low-pass filter and the sorting-based technique,
    algorithm in section 3.3.4 in [1]. Angular direction is along the axis 0.

    Parameters
    ---------
    sinogram : array_like
        2D array. Sinogram image.
    sigma : int
        Sigma of the Gaussian window.
    size : int
        Window size of the median filter.
    dim : {1, 2}, optional
        Dimension of the window.

    Returns
    -------
    ndarray
         2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1117/12.2530324
    """
    pad = min(150, int(0.1 * sinogram.shape[0]))
    sino_smooth = apply_gaussian_filter(sinogram, sigma, sigma, pad)
    sino_sharp = sinogram - sino_smooth
    sino_cor = remove_stripe_based_sorting(sino_sharp, size, dim)
    return sino_smooth + sino_cor


def remove_stripe_based_normalization(sinogram, sigma, num_chunk):
    """
    Remove stripes using the method in [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    sigma : int
        Sigma of the Gaussian window.
    num_chunk : int
        Number of chunks of rows.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://www.mcs.anl.gov/research/projects/X-ray-cmt/rivers/tutorial.html
    """
    (nrow, _) = sinogram.shape
    sinogram = np.copy(sinogram)
    list_index = np.array_split(np.arange(nrow), num_chunk)
    for pos in list_index:
        bindex = pos[0]
        eindex = pos[-1] + 1
        list_mean = np.mean(sinogram[bindex:eindex], axis=0)
        list_filt = gaussian_filter(list_mean, sigma)
        list_coe = list_filt - list_mean
        matcoe = np.tile(list_coe, (eindex - bindex, 1))
        sinogram[bindex:eindex, :] = sinogram[bindex:eindex, :] + matcoe
    return sinogram


def calculate_regularization_coefficient(width, alpha):
    """
    Calculate coefficients used for the regularization-based method.
    Eq. (7) in [1].

    Parameters
    ----------
    width : int
        Width of a square array.
    alpha : float
        Regularization parameter.

    Returns
    -------
    float
         Square array.

    References
    ----------
    .. [1] https://doi.org/10.1016/j.aml.2010.08.022
    """
    tau = 2.0 * np.arcsinh(np.sqrt(alpha) * 0.5)
    ilist = np.arange(0, width)
    jlist = np.arange(0, width)
    matjj, matii = np.meshgrid(jlist, ilist)
    mat1 = np.abs(matii - matjj)
    mat2 = matii + matjj
    mat1a = np.cosh((width - 1 - mat1) * tau)
    mat2a = np.cosh((width - mat2) * tau)
    matcoe = - (np.tanh(
        0.5 * tau) / (alpha * np.sinh(width * tau))) * (mat1a + mat2a)
    return matcoe


def remove_stripe_based_regularization(sinogram, alpha, num_chunk):
    """
    Remove stripes using the method in [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    alpha : float
        Regularization parameter, e.g 0.0005. Smaller is stronger
    num_chunk : int
        Number of chunks of rows.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1016/j.aml.2010.08.022
    """
    sinogram = -np.log(sinogram)
    (nrow, ncol) = sinogram.shape
    sijmat = calculate_regularization_coefficient(ncol, alpha)
    list_index = np.array_split(np.arange(nrow), num_chunk)
    list_grad = np.zeros(ncol, dtype=np.float32)
    mat_grad = np.zeros((ncol, ncol), dtype=np.float32)
    for pos in list_index:
        bindex = pos[0]
        eindex = pos[-1] + 1
        list_mean = np.mean(sinogram[bindex:eindex], axis=0)
        list_grad[1:-1] = (-1) * np.diff(list_mean, 2)
        list_grad[0] = list_mean[0] - list_mean[1]
        list_grad[-1] = list_mean[-1] - list_mean[-2]
        mat_grad[:] = list_grad
        list_coe = np.sum(mat_grad * sijmat, axis=1)
        mat_coe = np.tile(list_coe, (eindex - bindex, 1))
        sinogram[bindex:eindex, :] = sinogram[bindex:eindex, :] + mat_coe
    return np.exp(-sinogram)


def make_2d_butterworth_window(width, height, u, v, n):
    """
    Create a 2d window from the 1D Butterworth window.

    Parameters
    ----------
    height, width : int
        Shape of the window.
    u,n : int
        To define the shape of the 1D Butterworth window.
    v : int
        Number of rows (* 2) are the 1D Butterworth windows.

    Returns
    -------
    ndarray
        2D array.
    """
    xcenter = np.ceil(width / 2.0) - 1.0
    ycenter = np.int16(np.ceil(height / 2.0) - 1)
    listx = np.arange(width) - xcenter
    window = 1.0 / (1.0 + np.power(listx / u, 2 * n))
    row1 = ycenter - np.int16(v)
    row2 = ycenter + np.int16(v) + 1
    window_2d = np.ones((height, width), dtype=np.float32)
    window_2d[row1:row2] = window
    return window_2d


def remove_stripe_based_fft(sinogram, u, n, v, pad):
    """
    Remove stripes using the method in [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    u,n : int
        To define the shape of 1D Butterworth low-pass filter.
    v : int
        Number of rows (* 2) to be applied the filter.
    pad : int
        Padding for the Fourier transform.

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1063/1.1149043
    """
    if pad > 0:
        sinogram = np.pad(sinogram, ((pad, pad), (0, 0)), mode='mean')
        sinogram = np.pad(sinogram, ((0, 0), (pad, pad)), mode='edge')
    (nrow, ncol) = sinogram.shape
    window_2d = make_2d_butterworth_window(ncol, nrow, u, v, n)
    sinogram = fft.ifft2(
        np.fft.ifftshift(np.fft.fftshift(fft.fft2(sinogram)) * window_2d))
    return np.real(sinogram[pad:nrow - pad, pad:ncol - pad])


def remove_stripe_based_wavelet_fft(sinogram, level, sigma, order, pad):
    """
    Remove stripes using the method in [1].
    Angular direction is along the axis 0.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    level : int
        Wavelet decomposition level.
    sigma : int
        Damping parameter. Larger is stronger.
    order : int
        Order of the the Daubechies wavelets.
    pad : int
        Padding for FFT

    Returns
    -------
    ndarray
        2D array. Stripe-removed sinogram.

    Notes
    -----
    Code adapted from tomopy source code https://github.com/tomopy/tomopy
    with a small improvement of using different ways of padding to
    reduce the side effect of the Fourier transform.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.17.008567
    """
    (nrow, ncol) = sinogram.shape
    if pad > 0:
        sinogram = np.pad(sinogram, ((pad, pad), (0, 0)), mode='mean')
        sinogram = np.pad(sinogram, ((0, 0), (pad, pad)), mode='edge')
    # Wavelet decomposition
    cH = []
    cV = []
    cD = []
    wavelet_name = "db" + str(order)
    for j in range(level):
        sinogram, (cHt, cVt, cDt) = pywt.dwt2(sinogram, wavelet_name)
        cH.append(cHt)
        cV.append(cVt)
        cD.append(cDt)
    for j in range(level):
        fcV = np.fft.fftshift(np.fft.fft2(cV[j]))
        nrow1, ncol1 = fcV.shape
        y_hat = (np.arange(-nrow1, nrow1, 2, dtype=np.float32) + 1) / 2
        damp = 1 - np.exp(-np.power(y_hat, 2) / (2 * np.power(sigma, 2)))
        fcV = np.multiply(fcV, np.transpose(np.tile(damp, (ncol1, 1))))
        cV[j] = np.real(np.fft.ifft2(np.fft.ifftshift(fcV)))
    # Wavelet reconstruction.
    for j in range(level)[::-1]:
        sinogram = sinogram[0:cH[j].shape[0], 0:cH[j].shape[1]]
        sinogram = pywt.idwt2((sinogram, (cH[j], cV[j], cD[j])), wavelet_name)
    if pad > 0:
        sinogram = sinogram[pad:-pad, pad:-pad]
    return sinogram[:nrow, :ncol]


def remove_zinger(mat, threshold, size=1):
    """
    Remove zinger using the method in [1], working on a projection image or
    a singram image.

    Parameters
    ----------
    mat : array_like
        2D array. Projection image or sinogram image.
    threshold : float
        Threshold to segment zingers. Smaller is more sensitive.
        Recommended range [0.05, 0.1].
    size : int
        Size of a zinger.

    Returns
    -------
    ndarray
        2D array. Zinger-removed image.

    References
    ----------
    .. [1] N. T. Vo, R.C Atwood, M. Drakopoulos, and T. Connolley, "Data
    processing methods and data acquisition for samples larger than the
    field of view in parallel-beam tomography," Optics express, submitted.

    """
    step = np.clip(size, 1, None)
    mat = np.copy(mat)
    mat_ave = []
    for i in range(-size, size + 1, step):
        for j in range(-size, size + 1, step):
            if (i != 0) or (j != 0):
                mat_ave.append(np.roll(np.roll(mat, i, axis=0), j, axis=1))
    mat_ave = np.mean(np.asarray(mat_ave), axis=0)
    mat_ave[mat_ave == 0] = 1
    mat_nor = mat / mat_ave - 1.0
    mat[mat_nor > threshold] = mat_ave[mat_nor > threshold]
    return mat


def generate_blob_mask(flat, size, snr):
    """
    Generate a binary mask of blobs from a flat-field image [1].

    Parameters
    ----------
    flat : array_like
        2D array. Flat-field image.
    size : float
        Estimated size of the largest blob.
    snr : float
        Ratio used to segment blobs.

    Returns
    -------
    ndarray
        2D array. Binary mask.

    References
    ----------
    .. [1] N. T. Vo, R.C Atwood, M. Drakopoulos, and T. Connolley, "Data
    processing methods and data acquisition for samples larger than the
    field of view in parallel-beam tomography," Optics express, submitted.
    """
    mat = median_filter(flat, (2, 2))
    mask = np.zeros_like(mat)
    for i in range(mat.shape[0]):
        line = mat[i]
        line_fil = median_filter(line, size)
        line_fil[line_fil == 0.0] = np.mean(line_fil)
        line_norm = line / line_fil
        mask_1d = detect_stripe(line_norm, snr)
        mask_1d[0:2] = 0.0
        mask_1d[-2:] = 0.0
        mask[i] = np.float32(binary_dilation(mask_1d, iterations=1))
    return mask


def remove_blob_1d(sino_1d, mask_1d):
    """
    Remove blobs in one row of a sinogram.

    Parameters
    ----------
    sino_1d : array_like
        1D array. A row of a sinogram.
    mask_1d : array_like
        1D binary mask.

    Returns
    -------
    ndarray
        1D array.

    Notes
    -----
    The method is used to remove streak artifacts caused by blobs in
    a sinogram generated from a helical-scan data [1].

    References
    ----------
    .. [1] N. T. Vo, R.C Atwood, M. Drakopoulos, and T. Connolley, "Data
    processing methods and data acquisition for samples larger than the
    field of view in parallel-beam tomography," Optics express, submitted.
    """
    sino_1d = np.copy(sino_1d)
    listx = np.where(mask_1d < 1.0)[0]
    listy = sino_1d[listx]
    finter = interpolate.interp1d(listx, listy)
    mask_1d[:2] = 0.0
    mask_1d[-2:] = 0.0
    listx_miss = np.where(mask_1d > 0.0)[0]
    if len(listx_miss) > 0:
        sino_1d[listx_miss] = finter(listx_miss)
    return sino_1d


def remove_blob(mat, mask):
    """
    Remove blobs in an image.

    Parameters
    ----------
    mat : array_like
        2D array. Projection image or sinogram image.
    mask : array_like
        2D binary mask.

    Returns
    -------
    ndarray
        2D array.
    """
    mat = np.copy(mat)
    if mat.shape != mask.shape:
        raise ValueError("The image and the mask are not the same shape !!!")
    for i in range(mat.shape[0]):
        array_1d = mat[i]
        mask_1d = mask[i]
        mask_1d[:2] = 0.0
        mask_1d[-2:] = 0.0
        listx = np.where(mask_1d < 1.0)[0]
        listy = array_1d[listx]
        finter = interpolate.interp1d(listx, listy)
        listx_miss = np.where(mask_1d > 0.0)[0]
        if len(listx_miss) > 0:
            array_1d[listx_miss] = finter(listx_miss)
        mat[i] = array_1d
    return mat
