# ===========================================================================
# ===========================================================================
# Copyright (c) 2018 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===========================================================================
# Author: Nghia T. Vo
# E-mail: nghia.vo@diamond.ac.uk
# Description: Examples of how to use the algotom package to process a
# 360-degree scanning data with the offset center-of-rotation.
# ===========================================================================

"""
The following examples show how to use algotom to reconstruct a few slices
from a tomographic data acquired by using a 360-degree scan with the offset
center-of-rotation.

Raw data is at: https://zenodo.org/record/<>
There're 4 files: "projections_00000.hdf", "flats_00000.hdf", "darks_00000.hdf",
and "scan_00008.nxs" containing projection images, flat-field images, dark-field
images, and meta-data, respectively.

Referring to examples in "explore_hdf_tomo_data.py" to know how to find
key-paths and datasets in a hdf/nxs file.
"""

import numpy as np
import algotom.inputoutput.loadersaver as losa
import algotom.preprocessing.correction as corr
import algotom.preprocessing.removal as remo
import algotom.preprocessing.calculation as calc
import algotom.preprocessing.conversion as conv
import algotom.preprocessing.filtering as filt
import algotom.reconstruction.fftmethod as reco



proj_path = "C:/data/projections_00000.hdf"
flat_path = "C:/data/flats_00000.hdf"
dark_path = "C:/data/darks_00000.hdf"
meta_path = "C:/data/scan_00008.nxs"
key_path = "/entry/data/data"
angle_key = "/entry1/tomo_entry/data/rotation_angle"

output_base = "C:/output/"

data = losa.load_hdf(proj_path, key_path) # This is an object not ndarray.
(depth, height, width) = data.shape
angles = np.squeeze(np.asarray(losa.load_hdf(meta_path, angle_key)[:]))
# Load dark-field images and flat-field images, averaging each result.
flat_field = np.mean(losa.load_hdf(flat_path, key_path)[:], axis = 0)
dark_field = np.mean(losa.load_hdf(dark_path, key_path)[:], axis = 0)
# Generate a sinogram and perform reconstruction.
index = height // 2
print("1 -> Extract a sinogram with flat-field correction")
sino_360 = corr.flat_field_correction(data[:, index, :], flat_field[index], dark_field[index])
print("2 -> Calculate the center-of-rotation, the overlap-side and overlap-area used for stitching")
(center0, overlap, side,_) = calc.find_center_360(sino_360, 100)
print("Center-of-rotation: {0}. Side: {1} (0->'left', 1->'right'). Overlap: {2}".format(center0, side, overlap))
# 1st way: Convert the 360-degree sinogram to the 180-degree sinogram.
sino_180, center1 = conv.convert_sinogram_360_to_180(sino_360, center0)
losa.save_image(output_base + "/reconstruction/sino_180.tif", sino_180)

# Remove ring artifacts
sino_180 = remo.remove_all_stripe(sino_180, 3, 51, 17)

# Perform reconstruction
img_rec = reco.dfi_reconstruction(sino_180, center1, apply_log=True)
## Using gridrec code for faster reconstruction if Tomopy installed
# img_rec = reco.gridrec_reconstruction(sino_180, center1, apply_log=True)
losa.save_image(output_base + "/reconstruction/recon_image_1.tif", img_rec)


# 2nd way: Extending the 360-degree sinogram (by weighting and padding).
(sino_ext, center2) = conv.extend_sinogram(sino_360, center0)
losa.save_image(output_base + "/reconstruction/sino_360_extened.tif", sino_ext)
img_rec = reco.dfi_reconstruction(sino_ext, center2, angles=angles*np.pi/180.0,apply_log=False)
## Using fbp with GPU for faster reconstruction
# img_rec = reco.fbp_reconstruction(sino_ext, center2, angles=angles*np.pi/180.0,apply_log=False, gpu=True)
losa.save_image(output_base + "/reconstruction/recon_image_2.tif", img_rec)
print("!!! Done !!!")
